//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a tool.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------

using System;
using System.Reflection;

[assembly: System.Reflection.AssemblyCompanyAttribute("Python.Runtime")]
[assembly: System.Reflection.AssemblyConfigurationAttribute("release")]
[assembly: System.Reflection.AssemblyDescriptionAttribute("Python and CLR (.NET and Mono) cross-platform language interop")]
[assembly: System.Reflection.AssemblyFileVersionAttribute("3.0.0.0")]
[assembly: System.Reflection.AssemblyInformationalVersionAttribute("3.0.0-alpha2")]
[assembly: System.Reflection.AssemblyProductAttribute("Python.Runtime")]
[assembly: System.Reflection.AssemblyTitleAttribute("Python.Runtime")]
[assembly: System.Reflection.AssemblyVersionAttribute("3.0.0.0")]
[assembly: System.Reflection.AssemblyMetadataAttribute("RepositoryUrl", "https://github.com/pythonnet/pythonnet")]

// Von der MSBuild WriteCodeFragment-Klasse generiert.

